;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; MODULE      : init-equation-editor.scm
;; DESCRIPTION : plugin that enables using TeXmacs as a graphical
;;               equation editor for external programs 
;; COPYRIGHT   : (C) 2016  Philippe Joyez
;;
;; This software falls under the GNU general public license version 3 or later.
;; It comes WITHOUT ANY WARRANTY WHATSOEVER. For details, see the file LICENSE
;; in the root directory or <http://www.gnu.org/licenses/gpl-3.0.html>.
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; This plugin enables using TeXmacs as 
;; graphical "equation editor" which
;; can be called from other applications
;; (notably from within Inkscape using
;; the "Texmacs extension" which can be installed
;; in Inkscape from Texmacs convert->images preferences).
;; 
;; In this mode the equation to be edited 
;; is loaded from a temporary tm file and displayed with 
;; a toolbar with "done" and "cancel" buttons.
;; When the "done" button is pressed 
;; an updated svg image of the equation (containing the
;; texmacs representation of the equation) is created 
;;
;; The external application can communicate with texmacs 
;; (and this plugin) in two ways:
;; a) launching texmacs with command line 
;; options -x "(lazy-plugin-force)(equ-edit-cmdline)"
;; (see below) and using stdin/stdout pipes
;; b) connecting with an already running texmacs instance
;; using texmacs' socket server feature  
;; (enabled by the preference "equation-editor-server" 
;; controlled in the preference>convert>image widget).
;; Using socket communication is much faster since it
;; spares the boot-up time of texmacs.
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;(import-from (debug debug-menu))(debug-toggle "io")

(plugin-configure equation-editor 
  (:require #t)
  ;;(display "config equation-editor\n")
)

;;(when (supports-equation-editor?) (display "equation-editor plugin loaded\n"))


;; Do (server-start) automatically at boot-up if preferences are set so
;; to enable incoming connections

(if (get-boolean-preference "equation-editor-server")
  (begin 
    (import-from (server server-base)) ;; define tm-service for use below
    (with srv (client-start "localhost")
      (if (== srv -1) 
        (begin (display "starting server\n") (server-start))
        (begin (display "found local server\n")(client-stop srv)
        )))

;; need a login for connecting to texmac's server
    (server-set-user-information "inkscape" "inkscape-modify-equation" "inkscape" "" "no")

;; service where inkscape connects, sending the path of a temporary file
;; with the equation to be edited and (possibly) a latex code of the equation
;; (for compatibility with latex-based inkscape extensions)
    (tm-service (remote-equ file latex)
      (set! current-envelope envelope)
      (set! current-equ-url (system->url file))
        ;;(display* "remote-equ envelope = " current-envelope "\n")
        (load-buffer-in-new-window current-equ-url)
        (equ-edit-toolbar)
        (if (!= latex "") 
          (insert (latex->texmacs (parse-latex (string-append "\\[" latex "\\]")))))
    )
  ))

;; is operation in socket server mode preferred?
(define-preferences 
  ("equation-editor-server" "on" noop)) ;; actual pref value set in preference widget

(define (current-envelope) #f)
(define (current-equ-url) #f)

(tm-define (equ-edit-cmdline)
  (:synopsis "Shows equation-edit toolbar in non-server mode; can be called on command line with -x \"(lazy-plugin-force)(equ-edit-cmdline)\"")
  (set! current-envelope #f) ;;envelope shall be #f if not acting as socket server
  (set! current-equ-url (system->url (cAr (program-arguments))))
  (equ-edit-toolbar))

(define (equ-edit-done)
  (begin 
    (if (!= (current-buffer) current-equ-url) (switch-to-buffer current-equ-url))
    (select-all)
    (export-selection-as-graphics (url-glue current-equ-url ".svg"))
    (if (not current-envelope) (display "done\n") (server-return current-envelope #t) )
;; NB: when in non-server mode Inkscape watches for "done" in stdout           
    (equ-edit-end)))

(define (equ-edit-cancel)
  (begin 
    (if (not current-envelope ) (display "cancel\n") (server-return current-envelope  #f) )
    (equ-edit-end)))

(define (equ-edit-end) 
  (let ((stay (get-boolean-preference "equation-editor-server"))) ;; should we keep server running ?
;; we should also check (server-started?) but it's always false in qt-texmacs...
        (show-icon-bar 3 #f)
        (if (nnot current-envelope) (server-error current-envelope  "disconnect"))  ;;how to simply disconnect that client? 
        (if stay (begin   
          (if (> (length (window-list)) 1)  
              (begin (safely-kill-window) ))
          (if (== (length (buffer-list)) 1)  
              (begin (switch-to-buffer (buffer-new))))
          (buffer-close current-equ-url))
          (if (not current-envelope)  (quit) ))
  ))            

(define (equ-edit-toolbar)
  "Displays control buttons in the toolbar"
  (begin 
     
    (menu-bind texmacs-extra-icons 
       ((balloon "Done" "update equation in Inkscape") (equ-edit-done ) )
       ((balloon "Cancel" "abandon modifying equation") (equ-edit-cancel ))
       ((check (balloon "sansSerif" "") "v" (font-sserif?)) (toggle-sserif)))
    (show-icon-bar 3 #t)))

(define (font-sserif?) (== (get-init "font-family") "ss"))
(define (toggle-sserif) 
  (if (== (get-init "font-family") "ss") (init-default "font-family") (init-env "font-family" "ss")))


